.TH "Simple Virtual Machine - Plugin long. Towards infinite precision arithmetic" 7 "2022-01-30"
.SH SYNOPSIS
.B svm_plugin_long
This plugin gives access to numbers defined in the GMP library.
.SH DESCRIPTION
This plugin provides long number representation.
.P
Note: Only integers are currently supported, and real number may be added in the future.
.SH INVOKATION
This plugin can be added to the Simple Virtual Machine using one of these directives:
.SS Local
.nf

LOCAL PLUGIN "svmpluginlong/libsvmlong.so"

.fi
when compiled/installed locally and should be accessible from the application directory.
.SS Machine and System
.nf

PLUGIN "===PLUGINLIB==="

.fi
when installed with auto-tools and should be accessible from the Simple Virtual Machine installation directory or from system directories. (Recommended)
.SS Absolute path
.nf

PLUGIN "===PLUGINLIBDIR===/===PLUGINLIB==="

.fi
when installed with auto-tools and should be accessible from an absolute path.
.SH CONTENT
.SS TYPE long.int
The infinite precision integer.
.SS INSTRUCTION long.parse STR INT ?:base -> long.int
Safe string parsing for long.int conversion.
An optional base can be specified from 2 to 36.
When not given as parameter, the base 10 is used.
.TP
If the base or the string are incorrect, a NUMERIC interruption is raised.
.SS INSTRUCTION long.print long.int INT ?:base -> STR
Conversion to string.
An optional base can be specified from 2 to 36.
When not given as parameter, the base 10 is used.
.TP
If the base is incorrect, a NUMERIC interruption is raised.
.SS INSTRUCTION long.long INT -> long.int
Conversion from short integer.
.SS INSTRUCTION long.short long.int -> INT
Conversion to short integer.
.TP
Result is undefined when the long integer exceeds the capacity of a short integer.
.SS INSTRUCTION long.cmp [ INT long.int ] [ < > = <> <= => ] [ INT long.int ] -> BLN
Compares two integers, mixing short and long integers.
.SS INSTRUCTION long.sum [ INT long.int ] + -> long.int
Compute the sum of short and long integers.
.SS INSTRUCTION long.diff [ INT long.int ] 2 -> long.int
Compute the difference of short and long integers.
.SS INSTRUCTION long.prod [ INT long.int ] + -> long.int
Compute the product of short and long integers.
.SS INSTRUCTION long.quot [ INT long.int ] 2 -> long.int
Compute the quotient of short and long integers.
.TP
When the divisor is zero, a NUMERIC interruption is raised.
.SS INSTRUCTION long.rem [ INT long.int ] 2 -> long.int
Compute the remainder of short and long integers.
.TP
When the divisor is zero, a NUMERIC interruption is raised.
.SS SYSTEM INSTRUCTION long.exp [ INT long.int ] INT -> long.int
Compute the exponentiation for two integers.
The exposant is always a short integer.
.TP
When the exposant is negative, the instruction is unsafe.
.SS INSTRUCTION long.rand [ INT long.int ] -> long.int
Compute a pseudo-random number between zero and the limit given as parameter.
.SH EXAMPLE
.SS Factorial
.nf
#!===SVMBIN===
LOG
PLUGIN "svmcom.so"
PLUGIN "===PLUGINLIB==="
ARGUMENT STR n
PROCESS "main"
	CODE "main" INLINE
		:memory (PTR, long.int, long.int)/p
		:long.parse @&n -> (p/1)
		:call fact p
		:com.message @(p/2)
		:shutdown
	:label fact
		:memory BLN, PTR, long.int*2 -> &P
		:long.cmp @(P/1) = 0 -> &@&P
		:goto end_fact :when @&@&P TRUE
		:long.diff @(P/1) 1 -> (@&P/2)
		:call fact (@&P/1)*3
		:long.prod @(@&P/3) @(P/1) -> (P/2)
		:return
	:label end_fact
		CONST long.int "1" -> (P/2)
		:return
	END
	MEMORY n
END
.fi
.SH SEE ALSO
.BR svm (1)
for help on the Simple Virtual Machine.
.SH AUTHOR
This plugin has been written by Julien BRUGUIER and is maintained by Julien BRUGUIER <projet.svm@pappy.tf>.
