/*
 * Copyright (C) 2022-01-30  Julien BRUGUIER
 * 
 * Towards infinite precision arithmetic
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include <string>
#include <sstream>

#line 95 "../plugins/long/core/long.svm_plugin"
#include <gmp.h>
#line 27 "src/plugin.cpp"

#include <src/plugin.h>

extern "C"
{

void plugin_configure(void *plugin)
{
	::svm_plugin_configure(plugin,
		"PLUGIN long \n"
		"DEFINE \n"
		"	TYPE long.int \n"
		"	INSTRUCTION long.parse STR INT ?:base -> long.int \n"
		"	INSTRUCTION long.print long.int INT ?:base -> STR \n"
		"	INSTRUCTION long.long INT -> long.int \n"
		"	INSTRUCTION long.short long.int -> INT \n"
		"	INSTRUCTION long.cmp [ INT long.int ] [ < > = <> <= => ] [ INT long.int ] -> BLN \n"
		"	INSTRUCTION long.sum [ INT long.int ] + -> long.int \n"
		"	INSTRUCTION long.diff [ INT long.int ] 2 -> long.int \n"
		"	INSTRUCTION long.prod [ INT long.int ] + -> long.int \n"
		"	INSTRUCTION long.quot [ INT long.int ] 2 -> long.int \n"
		"	INSTRUCTION long.rem [ INT long.int ] 2 -> long.int \n"
		"	SYSTEM INSTRUCTION long.exp [ INT long.int ] INT -> long.int \n"
		"	INSTRUCTION long.rand [ INT long.int ] -> long.int \n"
		,SVM_API_SIGNATURE,SVM_VERSION);
}

#ifndef ARGV_VALUE
#define ARGV_VALUE(index,type) ::svm_value_##type##_get(svm,::svm_parameter_value_get(svm,argv[(index)]))
#endif
#ifndef ARGV_PLUGIN
#define ARGV_PLUGIN(index,plugin,name) reinterpret_cast<type_##name*>(::svm_value_plugin_get_internal(svm,::svm_parameter_value_get(svm,argv[(index)])))
#endif
#ifndef ARGV_MARKER
#define ARGV_MARKER(index) std::string(::svm_parameter_marker_get(svm,argv[(index)]).string)
#endif
#ifndef ARGV_KEYWORD
#define ARGV_KEYWORD(index) std::string(::svm_parameter_keyword_get(svm,argv[(index)]).string)
#endif
#ifndef ARGV_STRUCT
#define ARGV_STRUCT(index,plugin,name) reinterpret_cast<struct_##name*>(::svm_structure_get_internal(svm,::svm_value_pluginentrypoint_new__raw(svm,#plugin,#name),::svm_parameter_structure_get(svm,argv[(index)])))
#endif
#ifndef ARGV_VARIABLE
#define ARGV_VARIABLE(index) ::svm_parameter_variable_get(svm,argv[(index)]);
#endif
#ifndef NEW_VALUE
#define NEW_VALUE(type,value) ::svm_value_##type##_new(svm,(value))
#endif
#ifndef NEW_PLUGIN
#define NEW_PLUGIN(plugin,name,value) ::svm_value_plugin_new(svm,::svm_value_pluginentrypoint_new__raw(svm,#plugin,#name),(value))
#endif
#ifndef NEW_STRUCT
#define NEW_STRUCT(plugin,name,value) ::svm_structure_new(svm,::svm_value_pluginentrypoint_new__raw(svm,#plugin,#name),(value))
#endif
#ifndef NEW_STRING
#define NEW_STRING(raw_string) ::svm_string_new(svm,raw_string.c_str(),raw_string.size())
#endif
#ifndef RAW_STRING
#define RAW_STRING(svm_string) std::string(svm_string.string,svm_string.size)
#endif
#ifndef NEW_BOOLEAN
#define NEW_BOOLEAN(boolean) ((boolean)?TRUE:FALSE)
#endif
#ifndef RAW_BOOLEAN
#define RAW_BOOLEAN(boolean) ((boolean)==TRUE)
#endif
#ifndef NEW_NULL_VALUE
#define NEW_NULL_VALUE(type) ::svm_value_##type##_new_null(svm)
#endif
#ifndef NEW_NULL_PLUGIN
#define NEW_NULL_PLUGIN(plugin,name) ::svm_value_plugin_new_null(svm,::svm_value_pluginentrypoint_new__raw(svm,#plugin,#name))
#endif
#ifndef NEW_NULL_STRUCT
#define NEW_NULL_STRUCT(plugin,name) ::svm_structure_new_null(svm,::svm_value_pluginentrypoint_new__raw(svm,#plugin,#name))
#endif
#ifndef ERROR_INTERNAL
#define ERROR_INTERNAL(irq,message) ::svm_processor_current_raise_error_internal__raw(svm,irq,message)
#endif
#ifndef ERROR_EXTERNAL
#define ERROR_EXTERNAL(plugin,name,message) ::svm_processor_current_raise_error_external__raw(svm,::svm_value_pluginentrypoint_new__raw(svm,#plugin,#name),message)
#endif
#ifndef CONST_PEP
#define CONST_PEP(plugin,name) ::svm_value_pluginentrypoint_new__raw(svm,#plugin,#name)
#endif
#ifndef CURRENT
#define CURRENT(object) ::svm_##object##_get_current(svm)
#endif
#ifndef RETURN
#define RETURN return nullptr
#endif
#ifndef VARIABLE_GLOBAL
#define VARIABLE_GLOBAL(variable) ::svm_variable_scope_set_global(svm,(variable))
#endif
#ifndef VARIABLE_LOCAL
#define VARIABLE_LOCAL(variable) ::svm_variable_scope_set_local(svm,(variable))
#endif
#ifndef VARIABLE_DELETE
#define VARIABLE_DELETE(variable) ::svm_variable_delete(svm,(variable))
#endif

}

#line 99 "../plugins/long/core/long.svm_plugin"
static gmp_randstate_t state;
#line 132 "src/plugin.cpp"

extern "C"
{

/* TYPE long.int */

struct type_int
{
#line 114 "../plugins/long/core/long.svm_plugin"
	explicit type_int(const mpz_t& i)
	{
		::mpz_init_set(_int,i);
	}
	type_int(const type_int& i)
	{
		::mpz_init_set(_int,i._int);
	}
	~type_int()
	{
		::mpz_clear(_int);
	}
	mpz_t _int;
#line 155 "src/plugin.cpp"
};

void type_int_delete(const void *svm, void *handler)
{
	type_int * const object = reinterpret_cast<type_int*>(handler);
	{
#line 129 "../plugins/long/core/long.svm_plugin"

#line 164 "src/plugin.cpp"
	}
	delete object;
}

void* type_int_copy(const void *svm, const void *handler)
{
	const type_int *object = reinterpret_cast<const type_int*>(handler);
	type_int *copy = new type_int(*object);
	{
#line 130 "../plugins/long/core/long.svm_plugin"

#line 176 "src/plugin.cpp"
	}
	return copy;
}

void* type_int_constant(const void *svm, const SVM_String value)
{
	std::string string(value.string,value.size);
	{
#line 132 "../plugins/long/core/long.svm_plugin"
	mpz_t i;
	auto e = ::mpz_init_set_str(i,string.c_str(),10);
	if(e)
	{
		::mpz_set_si(i,0);
	}
	type_int *object = new type_int(i);
	return object;
#line 194 "src/plugin.cpp"
	}
}

SVM_String type_int_print(const void *svm, const void *handler)
{
	const type_int *object = reinterpret_cast<const type_int*>(handler);
	{
#line 143 "../plugins/long/core/long.svm_plugin"
	char *ss = ::mpz_get_str(nullptr,10,object->_int);
	std::string s = ss;
	::free(ss);
	return ::svm_string_new(svm,s.c_str(),s.size());
#line 207 "src/plugin.cpp"
	}
}


/* INSTRUCTION long.parse STR INT ?:base -> long.int */

SVM_Value instruction_parse(const void *svm, SVM_Size argc, SVM_Parameter argv[])
{
#line 152 "../plugins/long/core/long.svm_plugin"
	SVM_String integer = ARGV_VALUE(0,string);
	long long int base = 10;
	if(argc>1)
	{
		base = ::svm_value_integer_get(svm,::svm_parameter_value_get(svm,argv[1]));
		if((base<2) or (base>36))
		{
			ERROR_INTERNAL(NUMERIC,"Invalid base");
		}
	}
	mpz_t i;
	auto e = ::mpz_init_set_str(i,std::string(integer.string,integer.size).c_str(),base);
	if(e)
	{
		::mpz_clear(i);
		ERROR_INTERNAL(NUMERIC,"Invalid integer");
	}
	type_int *ii = new type_int(i);
	return NEW_PLUGIN(long,int,ii);
#line 236 "src/plugin.cpp"
}


/* INSTRUCTION long.print long.int INT ?:base -> STR */

SVM_Value instruction_print(const void *svm, SVM_Size argc, SVM_Parameter argv[])
{
#line 183 "../plugins/long/core/long.svm_plugin"
	type_int *integer = ARGV_PLUGIN(0,long,int);
	long long int base = 10;
	if(argc>1)
	{
		base = ::svm_value_integer_get(svm,::svm_parameter_value_get(svm,argv[1]));
		if((base<2) or (base>36))
		{
			ERROR_INTERNAL(NUMERIC,"Invalid base");
		}
	}
	char *ss = ::mpz_get_str(nullptr,base,integer->_int);
	std::string s = ss;
	::free(ss);
	return ::svm_value_string_new__buffer(svm,s.c_str(),s.size());
#line 259 "src/plugin.cpp"
}


/* INSTRUCTION long.long INT -> long.int */

SVM_Value instruction_long(const void *svm, SVM_Size argc, SVM_Parameter argv[])
{
#line 209 "../plugins/long/core/long.svm_plugin"
	long long int integer = ARGV_VALUE(0,integer);
	::mpz_t i;
	::mpz_init_set_si(i,integer);
	type_int *ii = new type_int(i);
	return NEW_PLUGIN(long,int,ii);
#line 273 "src/plugin.cpp"
}


/* INSTRUCTION long.short long.int -> INT */

SVM_Value instruction_short(const void *svm, SVM_Size argc, SVM_Parameter argv[])
{
#line 222 "../plugins/long/core/long.svm_plugin"
	type_int *integer = ARGV_PLUGIN(0,long,int);
	auto i = ::mpz_get_si(integer->_int);
	return NEW_VALUE(integer,i);
#line 285 "src/plugin.cpp"
}


/* INSTRUCTION long.cmp [ INT long.int ] [ < > = <> <= => ] [ INT long.int ] -> BLN */

SVM_Value instruction_cmp(const void *svm, SVM_Size argc, SVM_Parameter argv[])
{
#line 235 "../plugins/long/core/long.svm_plugin"
	mpz_t g;
	SVM_Value vg = ::svm_parameter_value_get(svm,argv[0]);
	if(::svm_value_type_is_integer(svm,vg))
	{
		auto gi = ::svm_value_integer_get(svm,vg);
		::mpz_init_set_si(g,gi);
	}
	else
	{
		type_int *integer = reinterpret_cast<type_int*>(::svm_value_plugin_get_internal(svm,vg));
		::mpz_init_set(g,integer->_int);
	}
	mpz_t d;
	SVM_Value vd = ::svm_parameter_value_get(svm,argv[2]);
	if(::svm_value_type_is_integer(svm,vd))
	{
		auto di = ::svm_value_integer_get(svm,vd);
		::mpz_init_set_si(d,di);
	}
	else
	{
		type_int *integer = reinterpret_cast<type_int*>(::svm_value_plugin_get_internal(svm,vd));
		::mpz_init_set(d,integer->_int);
	}
	SVM_String c = ::svm_parameter_marker_get(svm,argv[1]);
	std::string cmp(c.string,c.size);
	auto r = ::mpz_cmp(g,d);
	::mpz_clear(g);
	::mpz_clear(d);
	return ::svm_value_boolean_new__raw(svm,((r<0) and ((cmp=="<") or (cmp=="<>"))) or ((r<=0) and ((cmp=="<=") or (cmp=="<>"))) or ((r==0) and (cmp=="=")) or ((r>0) and ((cmp==">") or (cmp=="<>"))) or ((r>=0) and ((cmp=="=>") or (cmp=="<>"))));
#line 324 "src/plugin.cpp"
}


/* INSTRUCTION long.sum [ INT long.int ] + -> long.int */

SVM_Value instruction_sum(const void *svm, SVM_Size argc, SVM_Parameter argv[])
{
#line 273 "../plugins/long/core/long.svm_plugin"
	mpz_t s;
	::mpz_init_set_si(s,0);
	for(SVM_Index i = 0 ; i<argc ; ++i)
	{
		SVM_Value v = ::svm_parameter_value_get(svm,argv[i]);
		if(::svm_value_type_is_integer(svm,v))
		{
			auto d = ::svm_value_integer_get(svm,v);
			mpz_t vv;
			::mpz_init_set_si(vv,d);
			::mpz_add(s,s,vv);
			::mpz_clear(vv);
		}
		else
		{
			type_int *integer = reinterpret_cast<type_int*>(::svm_value_plugin_get_internal(svm,v));
			::mpz_add(s,s,integer->_int);
		}
	}
	type_int *ii = new type_int(s);
	return NEW_PLUGIN(long,int,ii);
#line 354 "src/plugin.cpp"
}


/* INSTRUCTION long.diff [ INT long.int ] 2 -> long.int */

SVM_Value instruction_diff(const void *svm, SVM_Size argc, SVM_Parameter argv[])
{
#line 299 "../plugins/long/core/long.svm_plugin"
	mpz_t g;
	SVM_Value vg = ::svm_parameter_value_get(svm,argv[0]);
	if(::svm_value_type_is_integer(svm,vg))
	{
		auto gi = ::svm_value_integer_get(svm,vg);
		::mpz_init_set_si(g,gi);
	}
	else
	{
		type_int *integer = reinterpret_cast<type_int*>(::svm_value_plugin_get_internal(svm,vg));
		::mpz_init_set(g,integer->_int);
	}
	mpz_t d;
	SVM_Value vd = ::svm_parameter_value_get(svm,argv[1]);
	if(::svm_value_type_is_integer(svm,vd))
	{
		auto di = ::svm_value_integer_get(svm,vd);
		::mpz_init_set_si(d,di);
	}
	else
	{
		type_int *integer = reinterpret_cast<type_int*>(::svm_value_plugin_get_internal(svm,vd));
		::mpz_init_set(d,integer->_int);
	}
	mpz_t r;
	::mpz_init(r);
	::mpz_sub(r,g,d);
	::mpz_clear(g);
	::mpz_clear(d);
	type_int *ii = new type_int(r);
	return NEW_PLUGIN(long,int,ii);
#line 394 "src/plugin.cpp"
}


/* INSTRUCTION long.prod [ INT long.int ] + -> long.int */

SVM_Value instruction_prod(const void *svm, SVM_Size argc, SVM_Parameter argv[])
{
#line 335 "../plugins/long/core/long.svm_plugin"
	mpz_t p;
	::mpz_init_set_si(p,1);
	for(SVM_Index i = 0 ; i<argc ; ++i)
	{
		SVM_Value v = ::svm_parameter_value_get(svm,argv[i]);
		if(::svm_value_type_is_integer(svm,v))
		{
			auto d = ::svm_value_integer_get(svm,v);
			::mpz_mul_si(p,p,d);
		}
		else
		{
			type_int *integer = reinterpret_cast<type_int*>(::svm_value_plugin_get_internal(svm,v));
			::mpz_mul(p,p,integer->_int);
		}
	}
	type_int *ii = new type_int(p);
	return NEW_PLUGIN(long,int,ii);
#line 421 "src/plugin.cpp"
}


/* INSTRUCTION long.quot [ INT long.int ] 2 -> long.int */

SVM_Value instruction_quot(const void *svm, SVM_Size argc, SVM_Parameter argv[])
{
#line 358 "../plugins/long/core/long.svm_plugin"
	mpz_t g;
	SVM_Value vg = ::svm_parameter_value_get(svm,argv[0]);
	if(::svm_value_type_is_integer(svm,vg))
	{
		auto gi = ::svm_value_integer_get(svm,vg);
		::mpz_init_set_si(g,gi);
	}
	else
	{
		type_int *integer = reinterpret_cast<type_int*>(::svm_value_plugin_get_internal(svm,vg));
		::mpz_init_set(g,integer->_int);
	}
	mpz_t d;
	SVM_Value vd = ::svm_parameter_value_get(svm,argv[1]);
	if(::svm_value_type_is_integer(svm,vd))
	{
		auto di = ::svm_value_integer_get(svm,vd);
		::mpz_init_set_si(d,di);
	}
	else
	{
		type_int *integer = reinterpret_cast<type_int*>(::svm_value_plugin_get_internal(svm,vd));
		::mpz_init_set(d,integer->_int);
	}
	if(mpz_sgn(d)==0)
	{
		::mpz_clear(g);
		::mpz_clear(d);
		ERROR_INTERNAL(NUMERIC,"Invalid divisor");
	}
	mpz_t r;
	::mpz_init(r);
	::mpz_tdiv_q(r,g,d);
	type_int *ii = new type_int(r);
	return NEW_PLUGIN(long,int,ii);
#line 465 "src/plugin.cpp"
}


/* INSTRUCTION long.rem [ INT long.int ] 2 -> long.int */

SVM_Value instruction_rem(const void *svm, SVM_Size argc, SVM_Parameter argv[])
{
#line 403 "../plugins/long/core/long.svm_plugin"
	mpz_t g;
	SVM_Value vg = ::svm_parameter_value_get(svm,argv[0]);
	if(::svm_value_type_is_integer(svm,vg))
	{
		auto gi = ::svm_value_integer_get(svm,vg);
		::mpz_init_set_si(g,gi);
	}
	else
	{
		type_int *integer = reinterpret_cast<type_int*>(::svm_value_plugin_get_internal(svm,vg));
		::mpz_init_set(g,integer->_int);
	}
	mpz_t d;
	SVM_Value vd = ::svm_parameter_value_get(svm,argv[1]);
	if(::svm_value_type_is_integer(svm,vd))
	{
		auto di = ::svm_value_integer_get(svm,vd);
		::mpz_init_set_si(d,di);
	}
	else
	{
		type_int *integer = reinterpret_cast<type_int*>(::svm_value_plugin_get_internal(svm,vd));
		::mpz_init_set(d,integer->_int);
	}
	if(mpz_sgn(d)==0)
	{
		::mpz_clear(g);
		::mpz_clear(d);
		ERROR_INTERNAL(NUMERIC,"Invalid divisor");
	}
	mpz_t r;
	::mpz_init(r);
	::mpz_tdiv_r(r,g,d);
	::mpz_clear(g);
	::mpz_clear(d);
	type_int *ii = new type_int(r);
	return NEW_PLUGIN(long,int,ii);
#line 511 "src/plugin.cpp"
}


/* SYSTEM INSTRUCTION long.exp [ INT long.int ] INT -> long.int */

SVM_Value instruction_exp(const void *svm, SVM_Size argc, SVM_Parameter argv[])
{
#line 450 "../plugins/long/core/long.svm_plugin"
	mpz_t g;
	SVM_Value vg = ::svm_parameter_value_get(svm,argv[0]);
	if(::svm_value_type_is_integer(svm,vg))
	{
		auto gi = ::svm_value_integer_get(svm,vg);
		::mpz_init_set_si(g,gi);
	}
	else
	{
		type_int *integer = reinterpret_cast<type_int*>(::svm_value_plugin_get_internal(svm,vg));
		::mpz_init_set(g,integer->_int);
	}
	auto di = ARGV_VALUE(1,integer);
	mpz_t r;
	::mpz_init(r);
	::mpz_pow_ui(r,g,di);
	::mpz_clear(g);
	type_int *ii = new type_int(r);
	return NEW_PLUGIN(long,int,ii);
#line 539 "src/plugin.cpp"
}


/* INSTRUCTION long.rand [ INT long.int ] -> long.int */

SVM_Value instruction_rand(const void *svm, SVM_Size argc, SVM_Parameter argv[])
{
#line 480 "../plugins/long/core/long.svm_plugin"
	mpz_t g;
	SVM_Value vg = ::svm_parameter_value_get(svm,argv[0]);
	if(::svm_value_type_is_integer(svm,vg))
	{
		auto gi = ::svm_value_integer_get(svm,vg);
		::mpz_init_set_si(g,gi);
	}
	else
	{
		type_int *integer = reinterpret_cast<type_int*>(::svm_value_plugin_get_internal(svm,vg));
		::mpz_init_set(g,integer->_int);
	}
	mpz_t r;
	::mpz_init(r);
	::mpz_urandomm(r,state,g);
	::mpz_clear(g);
	type_int *ii = new type_int(r);
	return NEW_PLUGIN(long,int,ii);
#line 566 "src/plugin.cpp"
}


/* Generic handling functions */

void plugin_initialisation(const void *svm)
{
#line 103 "../plugins/long/core/long.svm_plugin"
	::gmp_randinit_default(state);
#line 576 "src/plugin.cpp"
}

void plugin_finalisation(const void *svm)
{
#line 107 "../plugins/long/core/long.svm_plugin"
	::gmp_randclear(state);
#line 583 "src/plugin.cpp"
}

}
